@@ -24,18 +24,39 @@
 #define __GMIME_H__
 
 #include <glib.h>
+#include "gmime-charset.h"
+#include "gmime-iconv.h"
+#include "gmime-iconv-utils.h"
 #include "gmime-param.h"
 #include "gmime-content-type.h"
+#include "gmime-disposition.h"
+#include "gmime-data-wrapper.h"
+#include "gmime-object.h"
 #include "gmime-part.h"
 #include "gmime-message.h"
 #include "internet-address.h"
 #include "gmime-parser.h"
 #include "gmime-utils.h"
+#include "gmime-stream.h"
+#include "gmime-stream-buffer.h"
+#include "gmime-stream-file.h"
+#include "gmime-stream-filter.h"
+#include "gmime-stream-fs.h"
+#include "gmime-stream-mem.h"
+#include "gmime-stream-mmap.h"
+#include "gmime-stream-null.h"
+#include "gmime-filter.h"
+#include "gmime-filter-basic.h"
+#include "gmime-filter-charset.h"
+#include "gmime-filter-crlf.h"
+#include "gmime-filter-from.h"
+#include "gmime-filter-html.h"
+#include "gmime-filter-yenc.h"
 
 /* GMIME version */
-static const guint gmime_major_version = 0;
-static const guint gmime_minor_version = 6;
-static const guint gmime_micro_version = 0;
+static const guint gmime_major_version = 1;
+static const guint gmime_minor_version = 0;
+static const guint gmime_micro_version = 5;
 static const guint gmime_interface_age = 0;
 static const guint gmime_binary_age    = 0;
 
@@ -45,4 +66,14 @@ static const guint gmime_binary_age    = 0;
      (gmime_major_version == (major) && gmime_minor_version == (minor) && \
       gmime_micro_version >= (micro)))
 
+
+/**
+ * GMIME_INIT_FLAG_UTF8:
+ *
+ * Initialization flag to enable UTF-8 interfaces throughout GMime.
+ **/
+#define GMIME_INIT_FLAG_UTF8  (1 << 0)
+
+void g_mime_init (guint32 flags);
+
 #endif /* __GMIME_H__ */
