@@ -2,7 +2,7 @@
 /*
  *  Authors: Jeffrey Stedfast <fejj@ximian.com>
  *
- *  Copyright 2001 Ximain, Inc. (www.ximian.com)
+ *  Copyright 2001 Ximian, Inc. (www.ximian.com)
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -20,55 +20,63 @@
  *
  */
 
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
+#include <string.h>
+#include <ctype.h>
+#include <errno.h>
+
 #include "internet-address.h"
 #include "gmime-table-private.h"
 #include "gmime-utils.h"
-#include <string.h>
-#include <ctype.h>
+#include "gmime-iconv-utils.h"
+
 
 #define w(x) x
 
+
+extern int gmime_interfaces_utf8;
+
+
 /**
- * internet_address_new: Create a new Internet Address object
+ * internet_address_new:
+ *
+ * Creates a new InternetAddress object
  * 
  * Returns a new Internet Address object.
  **/
 InternetAddress *
 internet_address_new ()
 {
-	return g_new0 (InternetAddress, 1);
+	InternetAddress *ia;
+	
+	ia = g_new (InternetAddress, 1);
+	ia->type = INTERNET_ADDRESS_NONE;
+	ia->refcount = 1;
+	ia->name = NULL;
+	ia->value.addr = NULL;
+	
+	return ia;
 }
 
 
 /**
- * internet_address_destroy: Destroy an Internet Address object
- * @ia: Internet Address object to destroy
+ * internet_address_destroy:
+ * @ia: InternetAddress object to destroy
  * 
  * Destroy the InternetAddress object pointed to by @ia.
  **/
-void
+static void
 internet_address_destroy (InternetAddress *ia)
 {
 	if (ia) {
 		g_free (ia->name);
 		
 		if (ia->type == INTERNET_ADDRESS_GROUP) {
-			GList *members;
-			
-			members = ia->value.members;
-			while (members) {
-				InternetAddress *member;
-				
-				member = members->data;
-				internet_address_destroy (member);
-				members = members->next;
-			}
-			
-			g_list_free (ia->value.members);
+			internet_address_list_destroy (ia->value.members);
 		} else {
 			g_free (ia->value.addr);
 		}
@@ -79,9 +87,42 @@ internet_address_destroy (InternetAddress *ia)
 
 
 /**
- * internet_address_new_name: Create a new Internet Address object
+ * internet_address_ref:
+ * @ia: address
+ *
+ * Ref's the internet address.
+ **/
+void
+internet_address_ref (InternetAddress *ia)
+{
+	ia->refcount++;
+}
+
+
+/**
+ * internet_address_unref:
+ * @ia: address
+ *
+ * Unref's the internet address.
+ **/
+void
+internet_address_unref (InternetAddress *ia)
+{
+	if (ia->refcount <= 1) {
+		internet_address_destroy (ia);
+	} else {
+		ia->refcount--;
+	}
+}
+
+
+/**
+ * internet_address_new_name:
  * @name: person's name
  * @addr: person's address
+ *
+ * Creates a new InternetAddress object with name @name and address
+ * @addr.
  * 
  * Returns a new Internet Address object.
  **/
@@ -105,8 +146,10 @@ internet_address_new_name (const char *name, const char *addr)
 
 
 /**
- * internet_address_new_group: Create a new Internet Address object
+ * internet_address_new_group:
  * @name: group name
+ *
+ * Creates a new InternetAddress object with group name @name.
  * 
  * Returns a new Internet Address object.
  **/
@@ -174,24 +217,15 @@ internet_address_set_addr (InternetAddress *ia, const char *addr)
  * Set the members of the internet address group.
  **/
 void
-internet_address_set_group (InternetAddress *ia, GList *group)
+internet_address_set_group (InternetAddress *ia, InternetAddressList *group)
 {
-	GList *members;
+	/*InternetAddressList *members, *next; */
 	
 	g_return_if_fail (ia != NULL);
 	g_return_if_fail (ia->type != INTERNET_ADDRESS_NAME);
 	
 	ia->type = INTERNET_ADDRESS_GROUP;
-	members = ia->value.members;
-	while (members) {
-		InternetAddress *member;
-
-		member = members->data;
-		internet_address_destroy (member);
-		members = members->next;
-	}
-	
-	g_list_free (ia->value.members);
+	internet_address_list_destroy (ia->value.members);
 	ia->value.members = group;
 }
 
@@ -208,15 +242,166 @@ internet_address_add_member (InternetAddress *ia, InternetAddress *member)
 {
 	g_return_if_fail (ia != NULL);
 	g_return_if_fail (ia->type != INTERNET_ADDRESS_NAME);
+	g_return_if_fail (member != NULL);
 	
 	ia->type = INTERNET_ADDRESS_GROUP;
-	ia->value.members = g_list_append (ia->value.members, member);
+	ia->value.members = internet_address_list_append (ia->value.members, member);
 }
 
-static gchar *
-encoded_name (const gchar *raw, gboolean rfc2047_encode)
+
+/**
+ * internet_address_list_prepend:
+ * @list: a list of internet addresses
+ * @ia: internet address to prepend
+ *
+ * Prepends the internet address to the list of internet addresses
+ * pointed to by @list.
+ *
+ * Returns the resultant list.
+ **/
+InternetAddressList *
+internet_address_list_prepend (InternetAddressList *list, InternetAddress *ia)
 {
-	gchar *name;
+	InternetAddressList *node;
+	
+	g_return_val_if_fail (ia!=NULL, NULL);
+	
+	internet_address_ref (ia);
+	node = g_new (InternetAddressList, 1);
+	node->next = list;
+	node->address = ia;
+	
+	return node;
+}
+
+
+/**
+ * internet_address_list_append:
+ * @list: a list of internet addresses
+ * @ia: internet address to append
+ *
+ * Appends the internet address to the list of internet addresses
+ * pointed to by @list.
+ *
+ * Returns the resultant list.
+ **/
+InternetAddressList *
+internet_address_list_append (InternetAddressList *list, InternetAddress *ia)
+{
+	InternetAddressList *node, *n;
+	
+	g_return_val_if_fail (ia!=NULL, NULL);
+	
+	internet_address_ref (ia);
+	node = g_new (InternetAddressList, 1);
+	node->next = NULL;
+	node->address = ia;
+	
+	if (list == NULL)
+		return node;
+	
+	n = list;
+	while (n->next)
+		n = n->next;
+	
+	n->next = node;
+	
+	return list;
+}
+
+
+/**
+ * internet_address_list_concat:
+ * @a: first list
+ * @b: second list
+ *
+ * Concatenates a copy of list @b onto the end of list @a.
+ *
+ * Returns the resulting list.
+ **/
+InternetAddressList *
+internet_address_list_concat (InternetAddressList *a, InternetAddressList *b)
+{
+	InternetAddressList *node, *tail, *n;
+	
+	if (b == NULL)
+		return a;
+	
+	/* find the end of list a */
+	if (a != NULL) {
+		tail = a;
+		while (tail->next)
+			tail = tail->next;
+	} else {
+		tail = (InternetAddressList *) &a;
+	}
+	
+	/* concat a copy of list b to list a */
+	node = b;
+	while (node) {
+		internet_address_ref (node->address);
+		n = g_new (InternetAddressList, 1);
+		n->next = NULL;
+		n->address = node->address;
+		tail->next = n;
+		tail = n;
+		
+		node = node->next;
+	}
+	
+	return a;
+}
+
+
+/**
+ * internet_address_list_length:
+ * @list: list of internet addresses
+ *
+ * Calculates the length of the list of addresses.
+ *
+ * Returns the number of internet addresses in @list.
+ **/
+int
+internet_address_list_length (InternetAddressList *list)
+{
+	InternetAddressList *node;
+	int len = 0;
+	
+	node = list;
+	while (node) {
+		node = node->next;
+		len++;
+	}
+	
+	return len;
+}
+
+
+/**
+ * internet_address_list_destroy:
+ * @list: address list
+ *
+ * Destroys the list of internet addresses.
+ **/
+void
+internet_address_list_destroy (InternetAddressList *list)
+{
+	InternetAddressList *node, *next;
+	
+	node = list;
+	while (node) {
+		next = node->next;
+		internet_address_unref (node->address);
+		g_free (node);
+		node = next;
+	}
+}
+
+
+static char *
+encoded_name (const char *raw, gboolean rfc2047_encode)
+{
+	char *name;
 	
 	g_return_val_if_fail (raw != NULL, NULL);
 	
@@ -229,64 +414,95 @@ encoded_name (const gchar *raw, gboolean rfc2047_encode)
 	return name;
 }
 
+static void
+internet_address_list_to_string_internal (InternetAddressList *list, gboolean encode, GString *string)
+{
+	while (list) {
+		char *addr;
+		
+		addr = internet_address_to_string (list->address, encode);
+		if (addr) {
+			g_string_append (string, addr);
+			g_free (addr);
+			if (list->next)
+				g_string_append (string, ", ");
+		}
+		
+		list = list->next;
+	}
+}
+
 
 /**
- * internet_address_to_string: Write the InternetAddress object to a string
+ * internet_address_to_string:
  * @ia: Internet Address object
  * @encode: TRUE if the address should be rfc2047 encoded
+ *
+ * Allocates a string containing the contents of the InternetAddress
+ * object.
  * 
  * Returns the InternetAddress object as an allocated string in rfc822
  * format.
  **/
-gchar *
+char *
 internet_address_to_string (InternetAddress *ia, gboolean encode)
 {
-	char *string = NULL;
+	char *str = NULL;
 	
 	if (ia->type == INTERNET_ADDRESS_NAME) {
 		if (ia->name) {
 			char *name;
 			
 			name = encoded_name (ia->name, encode);
-			string = g_strdup_printf ("%s <%s>", name,
-						  ia->value.addr);
+			str = g_strdup_printf ("%s <%s>", name, ia->value.addr);
 			g_free (name);
 		} else {
-			string = g_strdup (ia->value.addr);
+			str = g_strdup (ia->value.addr);
 		}
 	} else if (ia->type == INTERNET_ADDRESS_GROUP) {
-		GList *members;
-		GString *gstr;
+		InternetAddressList *members;
+		GString *string;
 		
-		gstr = g_string_new (ia->name);
-		g_string_append (gstr, ": ");
+		string = g_string_new (ia->name);
+		g_string_append (string, ": ");
 		
 		members = ia->value.members;
-		while (members) {
-			InternetAddress *member;
-			char *addr;
-			
-			member = members->data;
-			members = members->next;
-			
-			addr = internet_address_to_string (member, encode);
-			if (addr) {
-				g_string_append (gstr, addr);
-				g_free (addr);
-				if (members)
-					g_string_append (gstr, ", ");
-			}
-		}
+		internet_address_list_to_string_internal (members, encode, string);
+		g_string_append (string, ";");
 		
-		g_string_append (gstr, ";");
-		
-		string = gstr->str;
-		g_string_free (gstr, FALSE);
+		str = string->str;
+		g_string_free (string, FALSE);
 	}
 	
-	return string;
+	return str;
 }
 
+
+/**
+ * internet_address_list_to_string:
+ * @list: list of internet addresses
+ * @encode: TRUE if the address should be rfc2047 encoded
+ *
+ * Allocates a string buffer containing the rfc822 formatted addresses
+ * in @list.
+ *
+ * Returns a string containing the list of addresses in rfc822 format.
+ **/
+char *
+internet_address_list_to_string (InternetAddressList *list, gboolean encode)
+{
+	GString *string;
+	char *str;
+	
+	string = g_string_new ("");
+	internet_address_list_to_string_internal (list, encode, string);
+	str = string->str;
+	g_string_free (string, FALSE);
+	
+	return str;
+}
+
+
 static void
 decode_lwsp (const char **in)
 {
@@ -347,48 +563,6 @@ decode_quoted_string (const char **in)
 	return out;
 }
 
-#if 0
-static char *
-decode_quoted_string (const char **in)
-{
-	const char *inptr = *in;
-	char *out = NULL, *outptr;
-	int outlen;
-	int c;
-	
-	decode_lwsp (&inptr);
-	if (*inptr == '"') {
-		const char *intmp;
-		int skip = 0;
-		
-		/* first, calc length */
-		inptr++;
-		intmp = inptr + 1;
-		while ((c = *intmp++) && c != '"') {
-			if (c == '\\' && *intmp) {
-				intmp++;
-				skip++;
-			}
-		}
-		
-		outlen = intmp - inptr - skip;
-		out = outptr = g_malloc (outlen + 1);
-		
-		while ((c = *inptr++) && c != '"') {
-			if (c == '\\' && *inptr) {
-				c = *inptr++;
-			}
-			*outptr++ = c;
-		}
-		*outptr = '\0';
-	}
-	
-	*in = inptr;
-	
-	return out;
-}
-#endif
-
 static char *
 decode_atom (const char **in)
 {
@@ -490,7 +664,7 @@ decode_domain (const char **in)
 				w(g_warning ("Missing ']' in domain-literal: %s", *in));
 		} else {
 			if (!(atom = decode_atom (&inptr))) {
-				w(g_warning ("Unexpeced char '%c' in domain: %s", *inptr, *in));
+				w(g_warning ("Unexpected char '%c' in domain: %s", *inptr, *in));
 				/* remove the last '.' */
 				if (domain->str[domain->len - 1] == '.')
 					g_string_truncate (domain, domain->len - 1);
@@ -584,6 +758,8 @@ decode_mailbox (const char **in)
 		if (name)
 			g_string_free (name, TRUE);
 		g_string_free (addr, TRUE);
+		*in = inptr;
+		
 		return NULL;
 	}
 	
@@ -654,8 +830,26 @@ decode_mailbox (const char **in)
 	
 	*in = inptr;
 	
-	if (addr->len)
+	if (addr->len) {
+		if (gmime_interfaces_utf8 && name && g_mime_utils_text_is_8bit (name->str, name->len)) {
+			/* A (broken) mailer has sent us an unencoded 8bit value.
+			 * Attempt to save it by assuming it's in the user's
+			 * locale and converting to UTF-8 */
+			char *buf;
+			
+			buf = g_mime_iconv_locale_to_utf8 (name->str);
+			if (buf) {
+				g_string_truncate (name, 0);
+				g_string_append (name, buf);
+				g_free (buf);
+			} else {
+				(g_warning ("Failed to convert \"%s\" to UTF-8: %s",
+					    name->str, g_strerror (errno)));
+			}
+		}
+		
 		mailbox = internet_address_new_name (name ? name->str : NULL, addr->str);
+	}
 	
 	g_string_free (addr, TRUE);
 	if (name)
@@ -690,6 +884,10 @@ decode_address (const char **in)
 	decode_lwsp (&inptr);
 	if (*inptr == ':') {
 		/* this is a group */
+		InternetAddressList *group = NULL, *tail;
+		
+		tail = (InternetAddressList *) &group;
+		
 		inptr++;
 		addr = internet_address_new_group (name->str);
 		
@@ -698,16 +896,24 @@ decode_address (const char **in)
 			InternetAddress *member;
 			
 			member = decode_mailbox (&inptr);
-			if (member)
-				internet_address_add_member (addr, member);
+			if (member) {
+				tail->next = g_new (InternetAddressList, 1);
+				tail = tail->next;
+				tail->next = NULL;
+				tail->address = member;
+			}
 			
 			decode_lwsp (&inptr);
 			while (*inptr == ',') {
 				inptr++;
 				decode_lwsp (&inptr);
 				member = decode_mailbox (&inptr);
-				if (member)
-					internet_address_add_member (addr, member);
+				if (member) {
+					tail->next = g_new (InternetAddressList, 1);
+					tail = tail->next;
+					tail->next = NULL;
+					tail->address = member;
+				}
 				
 				decode_lwsp (&inptr);
 			}
@@ -719,6 +925,8 @@ decode_address (const char **in)
 			w(g_warning ("Invalid group spec, missing closing ';': %.*s",
 				     inptr - start, start));
 		
+		internet_address_set_group (addr, group);
+		
 		*in = inptr;
 	} else {
 		/* this is a mailbox */
@@ -732,21 +940,24 @@ decode_address (const char **in)
 
 
 /**
- * internet_address_paarse_string:
+ * internet_address_parse_string:
  * @string: a string containing internet addresses
  *
  * Construct a list of internet addresses from the given string.
  *
- * Returns a linked list of internet addresses.
+ * Returns a linked list of internet addresses. *Must* be free'd by
+ * the caller.
  **/
-GList *
+InternetAddressList *
 internet_address_parse_string (const char *string)
 {
-	GList *addrlist = NULL;
+	InternetAddressList *node, *tail, *addrlist = NULL;
 	const char *inptr;
 	
 	inptr = string;
 	
+	tail = (InternetAddressList *) &addrlist;
+	
 	while (inptr && *inptr) {
 		InternetAddress *addr;
 		const char *start;
@@ -755,11 +966,16 @@ internet_address_parse_string (const char *string)
 		
 		addr = decode_address (&inptr);
 		
-		if (addr)
-			addrlist = g_list_append (addrlist, addr);
-		else
+		if (addr) {
+			node = g_new (InternetAddressList, 1);
+			node->next = NULL;
+			node->address = addr;
+			tail->next = node;
+			tail = node;
+		} else {
 			w(g_warning ("Invalid or incomplete address: %.*s",
 				     inptr - start, start));
+		}
 		
 		decode_lwsp (&inptr);
 		if (*inptr == ',')
